package risc;

import java.io.Reader;
import java.io.StreamTokenizer;
import java.io.IOException;
import java.util.Map;
import java.util.HashMap;
import java.util.ArrayList;

public class CodeParser {
    private HashMap/*<String, Int>*/ labels = new HashMap();
    private ArrayList/*<Instr>*/ instructions = new ArrayList();
    private int pos;

    //########################################################################

    public static int parse(int[] memory, int address, Reader reader) {
        return new CodeParser(reader).parseProgram(memory, address);
    }

    //########################################################################

    public static class Exception extends RuntimeException {

        public Exception(int line, String error) {
            super(error);
            this.line = line;
        }

        public String toString() {
            return "line " + line + ": " + getMessage();
        }

        public final int line;
    }

    //########################################################################

    abstract private class Instr {
        abstract public int encode(int address);
    };
    private class ResolvedInstr extends Instr {
        private final int opcode, a, b, c;
        public ResolvedInstr(int opcode, int a, int b, int c) {
            this.opcode = opcode;
            this.a = a;
            this.b = b;
            this.c = c;
        }
        public int encode(int address) {
            return RISC.Helper.code(opcode,a,b,c);
        }
    }
    private class UnresolvedInstr extends Instr {
        private final int opcode, a, b;
        private final String label;
        public UnresolvedInstr(int opcode, int a, int b, String label) {
            this.opcode = opcode;
            this.a = a;
            this.b = b;
            this.label = label;
        }
        public int encode(int address) {
            int c = ((Integer)labels.get(label)).intValue();
            switch (opcode) {
            case RISC.BEQ: case RISC.BNE:
            case RISC.BLT: case RISC.BGE:
            case RISC.BLE: case RISC.BGT:
            case RISC.BSR:
                c = (c / 4) - (address-1);
                break;
            case RISC.JSR:
                c = c / 4;
                break;
            default:
                // nothing to do.
            }
            return RISC.Helper.code(opcode, a, b, c);
        }
    }

    void addInstr(int op, int a, int b, int c) {
        instructions.add(new ResolvedInstr(op, a, b, c));
        pos += 4;
    }

    void addInstr(int op, int a, int b, String addressref) {
        instructions.add(new UnresolvedInstr(op, a, b, addressref));
        pos += 4;
    }

    private int parseProgram(int[] memory, int address) {
	pos = address;
        while (token.ttype!=token.TT_EOF) parseInstruction();
	for (int i = 0; i < instructions.size(); i++)
	    memory[address++]= ((Instr)instructions.get(i)).encode(address);

        return address;
    }

    private void parseInstruction() {
        if (token.ttype == token.TT_WORD) {
            String mnemonic = token.sval;

            if (mnemonic.endsWith(":")) {
                // Label definition
		labels.put(mnemonic.substring(0, mnemonic.length() - 1),
                           new Integer(pos));
		nextToken();
	    } else if (mnemonic.equals("call")) {
                // Pseudo-instruction: call
                nextToken();
                addInstr(RISC.ORI, 31, 0, pos + 8);
                addInstr(RISC.RET, 0, 0, parseReg());
            } else if (mnemonic.equals("lda")) {
                // Pseudo-instruction: lda
                nextToken();
                addInstr(RISC.ORI, parseReg(), 0, parseLabel());
            } else {
                // Normal instruction.
                Object object = opcodes.get(mnemonic);
                if (object == null)
                    throw error("unknown opcode " + mnemonic);
                nextToken();
                int opcode = ((Integer)object).intValue();
                int format = RISC.opcode_format[opcode];

                switch (format) {
                case RISC.OpRaRbRc:
                    addInstr(opcode, parseReg(), parseReg(), parseReg());
                    break;
                case RISC.OpRaRbIc: {
                    int a = parseReg();
                    int b = parseReg();
                    if (token.ttype == token.TT_WORD)
                        addInstr(opcode, a, b, parseLabel());
                    else
                        addInstr(opcode, a, b, parseSigned(16));
                } break;
                case RISC.OpRaRbUc: {
                    int a = parseReg();
                    int b = parseReg();
                    if (token.ttype == token.TT_WORD)
                        addInstr(opcode, a, b, parseLabel());
                    else
                        addInstr(opcode, a, b, parseUnsigned(16));
                } break;
                case RISC.OpRa__Rc:
                    addInstr(opcode, parseReg(), 0, parseReg());
                    break;
                case RISC.OpRa__Ic: {
                    int a = parseReg();
                    if (token.ttype == token.TT_WORD)
                        addInstr(opcode, a, 0, parseLabel());
                    else
                        addInstr(opcode, a, 0, parseSigned(16));
                } break;
                case RISC.OpRa__Uc: {
                    int a = parseReg();
                    if (token.ttype == token.TT_WORD)
                        addInstr(opcode, a, 0, parseLabel());
                    else
                        addInstr(opcode, a, 0, parseUnsigned(16));
                } break;
                case RISC.OpRa__Oc: {
                    int a = parseReg();
                    if (token.ttype == token.TT_WORD)
                        addInstr(opcode, a, 0, parseLabel());
                    else
                        addInstr(opcode, a, 0, parseSigned(21));
                } break;
                case RISC.OpRa____:
                    addInstr(opcode, parseReg(), 0, 0);
                    break;
                case RISC.Op____Rc:
                    addInstr(opcode, 0, 0, parseReg());
                    break;
                case RISC.Op____Oc: {
                    if (token.ttype == token.TT_WORD)
                        addInstr(opcode, 0, 0, parseLabel());
                    else
                        addInstr(opcode, 0, 0, parseSigned(21));
                } break;
                case RISC.Op____Lc: {
                    if (token.ttype == token.TT_WORD)
                        addInstr(opcode, 0, 0, parseLabel());
                    else
                        addInstr(opcode, 0, 0, parseUnsigned(26));
                } break;
                case RISC.Op______:
                    addInstr(opcode, 0, 0, 0);
                    break;
                default:
                    throw new Error("unknown instruction format " + format);
                }
            }
        } else throw error("opcode expected found " + token);
    }

    //########################################################################

    private int parseReg() {
        if (token.ttype == token.TT_WORD) {
	    String reg = token.sval;
	    if (reg.startsWith("r")) {
		int register = Integer.parseInt(reg.substring(1));
		if (0 <= register && register < 32) {
		    nextToken();
		    return register;
		}
		throw error("wrong register number: " + register);
	    }
	    throw error("expected a register, found " + token);
        } else if (token.ttype == token.TT_NUMBER) {
	    int register = (int)token.nval;
	    if (0 <= register && register < 32) {
		nextToken();
		return register;
	    }
	    throw error("wrong register number: " + register);
	}
        throw error("expected a register, found " + token);
    }

    private String parseLabel() {
        if (token.ttype == token.TT_WORD) {
            String val = token.sval;
            nextToken();
            return val;
        } else
            throw error("expected a label, found " + token);
    }

    private int parseSigned(int nbits) {
        if (token.ttype == token.TT_NUMBER) {
            int value = (int)token.nval;
            if ((-1 << (nbits - 1)) <= value && value < (1 << (nbits - 1))) {
                nextToken();
                return value & ((1 << nbits) - 1);
            }
            throw error("signed immediate field too large: " + value);
        }
        throw error("expected a signed immediate field, found " + token);
    }

    private int parseUnsigned(int nbits) {
        if (token.ttype == token.TT_NUMBER) {
            int value = (int)token.nval;
            if (value < (1 << nbits)) {
                nextToken();
                return value & ((1 << nbits) - 1);
            }
            throw error("unsigned immediate field too large: " + value);
        }
        throw error("expected an unsigned immediate field, found " + token);
    }

    //########################################################################

    private void nextToken() {
        try {
            token.nextToken();
        } catch (IOException exception) {
            throw error(exception.toString());
        }
    }

    private Exception error(String message) {
        return new Exception(token.lineno(), message);
    }

    //########################################################################

    private CodeParser(Reader stream) {
        this.token = new StreamTokenizer(stream);
        this.opcodes = RISC.Helper.getMnemonicToOpcodeMap();
        token.parseNumbers();
	token.whitespaceChars(',',',');
        token.slashSlashComments(true);
        token.slashStarComments(true);
        token.eolIsSignificant(false);
        token.lowerCaseMode(true);
	token.wordChars(':','_');
        nextToken();
    }

    private final StreamTokenizer token;
    private final Map opcodes;

    //########################################################################
}
