package risc.debugger;

import java.io.Reader;
import java.io.Writer;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;

import risc.RISC;
import risc.emulator.Emulator;
import risc.emulator.SysCallHandler;

public class Debugger extends Emulator {

    //########################################################################
    // Private state

    private final Stack trace;
    private final Map bps;
    private boolean enabled;

    //########################################################################
    // Constructors

    public Debugger(int memory_size, SysCallHandler os) {
        super(memory_size, os);
        this.trace = new Stack();
        this.bps = new HashMap();
        this.enabled = true;
    }

    //########################################################################
    // Debugger interface - !!!

    public int getTraceSize() {
        return trace.size();
    }

    public Function getTraceAt(int index) {
        return (Function)trace.elementAt(index);
    }

    public Function[] getTrace() {
        return (Function[])trace.toArray(new Function[trace.size()]);
    }

    //########################################################################
    // Debugger interface - !!!

    public int getCodeAt(int adr) {
        if (enabled) {
            Integer code = (Integer)bps.get(new Integer(adr));
            if (code != null) return code.intValue();
        }
        return getWordAt(adr);
    }

    public int setCodeAt(int adr, int value) {
        Integer code = (Integer)bps.get(new Integer(adr));
        if (code != null) {
            bps.put(new Integer(adr), new Integer(value));
            if (enabled) return value;
        }
        return setWordAt(adr, value);
    }

    //########################################################################
    // Debugger interface - !!!

    public boolean isBP_(int adr) {
        return bps.containsKey(new Integer(adr));
    }

    public void toggleBP_(int adr) { // !!! remove ?
        if (isBP_(adr)) removeBP_(adr); else insertBP_(adr);
    }

    public void insertBP_(int adr) {
        // !!! check adr
        if (isBP_(adr)) return;
        bps.put(new Integer(adr), new Integer(getWordAt(adr)));
        if (enabled) setWordAt(adr, RISC.BREAK << 26); // !!! 
    }

    public void removeBP_(int adr) {
        Integer code = (Integer)bps.remove(new Integer(adr));
        if (enabled && code != null) setWordAt(adr, code.intValue());
    }

    public void enableBPs(boolean enable) {
        if (enable == enabled) return;
        Iterator entries = bps.entrySet().iterator();
        while (entries.hasNext()) {
            Map.Entry entry = (Map.Entry)entries.next();
            int adr = ((Integer)entry.getKey()).intValue();
            Integer code = (Integer)entry.getValue();
            if (enable) {
                setWordAt(adr, RISC.BREAK << 26); // !!! 
            } else {
                setWordAt(adr, code.intValue());
            }
        }
        enabled = enable;
    }

    public void removeBPs() {
        bps.clear();
    }

    //########################################################################
    



    //########################################################################



    //########################################################################
    //########################################################################
    //########################################################################
    //########################################################################

    //########################################################################
    // Debugger interface - execution

    public void run(int mintrace) throws Emulator.Exception {
        while (trace.size() >= mintrace) step();
    }

    // !!! check
    public void runThrough() throws Emulator.Exception {
        boolean enabled = this.enabled;
        if (enabled) enableBPs(false);
        run();
        if (enabled) enableBPs(true);
    }

    public void stepIn() throws Emulator.Exception {
        step();
    }

    public void stepOver() throws Emulator.Exception {
        int mintrace = getTraceSize() + 1;
        step();
	run(mintrace);
    }

    public void stepOut() throws Emulator.Exception {
        run(getTraceSize());
    }

    //########################################################################
    // Emulator interface - fetch

    protected Void call(int adr) throws Emulator.Exception {
	String name = "F" + (trace.size() + 1);
	trace.push(new Function(4 * adr, name, getRxs(), getPC()));
        return super.call(adr);
    }

    protected Void ret(int adr) throws Emulator.Exception {
        if (!trace.isEmpty()) trace.pop();
        return super.ret(adr);
    }

    protected Void breakpoint() throws Emulator.Exception {
        Integer instr = (Integer)bps.get(new Integer(getPC()));
        if (instr != null) setIR(instr.intValue());
        return super.breakpoint();
    }

    //########################################################################
}
