package risc.emulator;

/** !!!
 * Risc emulator.<br>
 * Java implementation of the emulator of the simple RISC processor
 * described in:<br>
 * N. Wirth, <i>Compiler Construction</i>, Addison Wesley 1996.
 *
 * @author Benjamin Leroy-Beaulieu
 *
 * Highly inspired from the Matteo Corti RISC processor emulator.
 */

// !!! check
import java.io.Reader;

import risc.RISC;
import risc.CodeWriter;
import risc.CodeParser;
import risc.StringHelper;

public class Emulator {
    public static final int DEFAULT_MEMORY_SIZE = 16384;

    //########################################################################
    // Private state

    private final SysCallHandler os;
    private final int[] m;
    private final int[] r;

    private int pc;
    private int ir;

    private long[] instructionCounter =  new long[RISC.opcode_count];

    private volatile boolean stop = false;

    //########################################################################
    // Exceptions

    public static class Exception extends /*Runtime*/java.lang.Exception {
        public Exception() {
            super();
        }
        public Exception(String message) {
            super(message);
        }
    }
    public static class Error extends Exception {
        public Error(String error) {
            super(error);
        }
        public Error(String error, Throwable exception) {
            this(error(error, exception));
        }
        public static String error(String error, Throwable exception) {
            String message = exception != null ? exception.getMessage() : null;
            return error + (message != null ? ": " + message : "");
        }
    }
    public static class Break extends Exception {}
    public static class Interrupt extends Exception {}
    public static class Exit extends Exception {
        public Exit(int code) {
            this.code = code;
        }
        public String toString() {
            return "Program exited with code " + code;
        }
        public final int code;
    }

    //########################################################################
    // Constructors

    public Emulator(int memory_size, SysCallHandler os) {
        if (memory_size < 0)
            throw illegal("invalid memory size: " + memory_size);
        if ((memory_size & 0x03) != 0)
            throw illegal("nonaligned memory size: " + memory_size);
        if (os == null)
            throw illegal("null os");
        this.os = os;
        this.m = new int[memory_size / 4];
        this.r = new int[RISC.register_count];
        this.reset();
    }

    //########################################################################
    // Emulator interface - state

    public void reset() {
        for (int i = 0; i < m.length; i++) m[i] = 0;
        for (int i = 0; i < r.length; i++) r[i] = 0;
        pc = -1;;
        ir = 0;
        stop = false;
    }

    public int getMemorySize() {
        return 4 * m.length;
    }

    public byte getByteAt(int adr) {
        if (adr < 0 || adr >= 4 * m.length)
            throw illegal("out of bounds address: " + toHexString(adr));
        return (byte)(m[adr / 4] >> (8 * (adr % 4)));
    }

    public int getWordAt(int adr) {
        if (adr < 0 || adr >= 4 * m.length)
            throw illegal("out of bounds address: " + toHexString(adr));
        if (adr % 4 != 0)
            throw illegal("nonaligned address: " + toHexString(adr));
        return m[adr / 4];
    }

    public byte setByteAt(int adr, byte value) {
        if (adr < 0 || adr >= 4 * m.length)
            throw illegal("out of bounds address: " + toHexString(adr));
        int shift = 8 * (adr % 4);
        m[adr / 4] = (m[adr / 4] & ~(0xFF << shift)) | (value << shift);
        return value;
    }

    public int setWordAt(int adr, int value) {
        if (adr < 0 || adr >= 4 * m.length)
            throw illegal("out of bounds address: " + toHexString(adr));
        if (adr % 4 != 0)
            throw illegal("nonaligned address: " + toHexString(adr));
        return m[adr / 4] = value;
    }

    public int getRx(int x) {
        if (x < 0 || x >= r.length) throw illegal("unknown register: " + x);
        return x != 0 ? r[x] : 0;
    }

    public int[] getRxs() { // !!!
        int[] newr = new int[r.length];
        System.arraycopy(r, 0, newr, 0, r.length);
        return newr;
    }

    public int setRx(int x, int value) {
        if (x < 0 || x >= r.length) throw illegal("unknown register: " + x);
        return x != 0 ? r[x] = value : 0;
    }

    public int getPC() {
        return pc * 4;
    }

    public int setPC(int adr) {
        if (adr % 4 != 0)
            throw illegal("nonaligned address: " + toHexString(adr));
        pc = adr / 4;
        return adr;
    }

    public int getIR() {
        return ir;
    }

    public int setIR(int value) {
        return ir = value;
    }

    public long[] getInstructionCounters() {
        return instructionCounter;
    }

    // !!! remove ?
    public int assemble(Reader reader) {
        // !!! try {
        return CodeParser.parse(m, 0, reader) * 4;
        // !!! } catch (IOException e) {
        // !!! startgui.printMessage("I/O error");
        // !!! }
    }

    //########################################################################
    // Emulator interface - execution

    public void run() throws Emulator.Exception {
        while (true) step();
    }

    // !!! remove ?
    public int syscall(int sc, int a, int b) throws Emulator.Exception {
        return os.handle(sc, a, b, this);
    }

    // !!! remove ?
    public int syscall_nofail(int sc, int a, int b) {
        try {
            return os.handle(sc, a, b, this);
        } catch (Emulator.Exception exception) {
            return 0;
        }
    }

    public void stop() {
        stop = true;
    }

    //########################################################################
    // Emulator interface - fetch

    protected Void error(String error) throws Emulator.Exception {
        return error(error, null);
    }

    protected Void error(String error, Throwable exception)
        throws Emulator.Exception {
        throw new Emulator.Error(error, exception);
    }

    protected Void exit() throws Emulator.Exception {
        pc = -1;
        ir = 0;
        throw new Emulator.Exit(0);
    }

    protected Void fetch(int adr) throws Emulator.Exception {
        if (adr >= m.length || adr < 0)
            return error("wrong instruction address: " + toHexString(4*adr));
        ir = m[pc = adr];
        return null;
    }

    protected Void next() throws Emulator.Exception {
        return fetch(pc + 1);
    }

    protected Void branch(int adr) throws Emulator.Exception {
        return fetch(adr);
    }

    protected Void call(int adr) throws Emulator.Exception {
        r[RISC.LNK] = 4 * (pc + 1);
        return fetch(adr);
    }

    protected Void ret(int adr) throws Emulator.Exception {
        return adr == 0 ? exit() : fetch(adr);
    }

    protected Void breakpoint() throws Emulator.Exception {
        throw new Emulator.Break();
    }

    protected Void interrupt() throws Emulator.Interrupt {
        throw new Emulator.Interrupt();
    }

    //########################################################################
    // Emulator interface - step

    protected Void step() throws Emulator.Exception {
        if (stop) { stop = false; return interrupt(); }
        r[0] = 0;

        int opcode = RISC.Helper.op(ir);
        int format = RISC.opcode_format[opcode];
        int a = RISC.Helper.ra(ir);
        int b = RISC.Helper.rb(ir);
        int c = 0;
        switch (format) {
        case RISC.OpRaRbRc:
        case RISC.OpRa__Rc:
        case RISC.Op____Rc: c = r[RISC.Helper.rc(ir)]; break;
        case RISC.OpRaRbIc:
        case RISC.OpRa__Ic: c =   RISC.Helper.ic(ir) ; break;
        case RISC.OpRaRbUc:
        case RISC.OpRa__Uc: c =   RISC.Helper.uc(ir) ; break;
        case RISC.OpRa__Oc:
        case RISC.Op____Oc: c =   RISC.Helper.oc(ir) ; break;
        case RISC.OpRa____:
        case RISC.Op______:                                    break;
        case RISC.Op____Lc: c =   RISC.Helper.lc(ir) ; break;
        default: throw new Error("unknown instruction format " + format);
        }

        instructionCounter[opcode]++;

        switch (opcode) {

        case RISC.ADD:
        case RISC.ADDI:
        case RISC.ADDIU:
            r[a] = r[b] + c;
            return next();

        case RISC.SUB:
        case RISC.SUBI:
        case RISC.SUBIU:
            r[a] = r[b] - c;
            return next();

        case RISC.CMP:
        case RISC.CMPI:
        case RISC.CMPIU:
            // In case of overflow the result of CMPx has the wrong
            // magnitude but the correct sign.
            b = r[b];
            if ((c < 0) && (b > Integer.MAX_VALUE + c))
                r[a] = Integer.MAX_VALUE;
            else if ((c > 0) && (b < Integer.MIN_VALUE + c))
                r[a] = Integer.MIN_VALUE;
            else
                r[a] = b - c;
            return next();

        case RISC.MUL:
        case RISC.MULI:
        case RISC.MULIU:
            r[a] = r[b] * c;
            return next();

        case RISC.DIV:
        case RISC.DIVI:
        case RISC.DIVIU:
            if (c == 0) return error("division by 0: " + b + " / " + c);
            r[a] = r[b] / c;
            return next();

        case RISC.MOD:
        case RISC.MODI:
        case RISC.MODIU:
            if (c == 0) return error("division by 0: " + b + " % " + c);
            r[a] = r[b] % c;
            return next();

        case RISC.OR:
        case RISC.ORI:
        case RISC.ORIU:
            r[a] = r[b] | c;
            return next();

        case RISC.AND:
        case RISC.ANDI:
        case RISC.ANDIU:
            r[a] = r[b] & c;
            return next();

        case RISC.BIC:
        case RISC.BICI:
        case RISC.BICIU:
            r[a] = r[b] & ~ c;
            return next();

        case RISC.XOR:
        case RISC.XORI:
        case RISC.XORIU:
            r[a] = r[b] ^ c;
            return next();

        case RISC.LSH:
        case RISC.LSHI:
            r[a] = c > 0 ? r[b] << c : r[b] >>> -c;
            return next();

        case RISC.ASH:
        case RISC.ASHI:
            r[a] = c > 0 ? r[b] << c : r[b] >> -c;
            return next();

        case RISC.CHK:
        case RISC.CHKI:
        case RISC.CHKIU:
            a = r[a];
            if ((a < 0) || (a >= c))
                return error("check failed: 0 <= " + c + " < " + a);
            return next();

        case RISC.LDW:
            c = r[b] + c;
            if (c >= 4 * m.length || c < 0)
                return error("wrong value address: " + toHexString(c));
            r[a] = m[c / 4];
            return next();

        case RISC.LDB:
            c = r[b] + c;
            if (c >= 4 * m.length || c < 0)
                return error("wrong value address: " + toHexString(c));
            r[a] = (m[c / 4] << (3 - (c % 4)) * 8) >>> 24;
            return next();

        case RISC.POP:
            if (r[b] >= 4 * m.length || r[b] < 0)
                return error("wrong value address: " + toHexString(r[b]));
            r[a] = m[r[b] / 4];
            r[b]+= c;
            return next();

        case RISC.STW:
            c = r[b] + c;
            if (c >= 4 * m.length || c < 0)
                return error("wrong value address: " + toHexString(c));
            m[c / 4] = r[a];
            return next();

        case RISC.STB:
            c = r[b] + c;
            if (c >= 4 * m.length || c < 0)
                return error("wrong value address: " + toHexString(c));
            b = (c % 4) * 8;
            m[c / 4] = (~(0xFF << b) & m[c / 4]) | (r[a] << b);
            return next();

        case RISC.PSH:
            c = r[b] - c;
            if (c >= 4 * m.length || c < 0)
                return error("wrong value address: " + toHexString(c));
            m[(r[b] = c) / 4] = r[a];
            return next();

        case RISC.BEQ:
            return r[a] == 0 ? branch(pc + c) : next();

        case RISC.BNE:
            return r[a] != 0 ? branch(pc + c) : next();

        case RISC.BLT:
            return r[a] < 0 ? branch(pc + c) : next();

        case RISC.BGE:
            return r[a] >= 0 ? branch(pc + c) : next();

        case RISC.BLE:
            return r[a] <= 0 ? branch(pc + c) : next();

        case RISC.BGT:
            return r[a] > 0 ? branch(pc + c) : next();

        case RISC.BSR:
            return call(pc + c);

        case RISC.JSR:
            return call(c);

        case RISC.RET:
            return ret(c / 4);

        case RISC.RD:
            syscall(RISC.SYS_IO_WR_CHR, 'i', 0);
            syscall(RISC.SYS_IO_WR_CHR, 'n', 0);
            syscall(RISC.SYS_IO_WR_CHR, 'p', 0);
            syscall(RISC.SYS_IO_WR_CHR, 'u', 0);
            syscall(RISC.SYS_IO_WR_CHR, 't', 0);
            syscall(RISC.SYS_IO_WR_CHR, ':', 0);
            syscall(RISC.SYS_IO_WR_CHR, ' ', 0);
            syscall(RISC.SYS_IO_FLUSH, 0, 0);
            r[a] = syscall(RISC.SYS_IO_RD_INT, 0, 0);
            return next();

        case RISC.WRD:
            syscall(RISC.SYS_DB_WR_REG, r[a], 0);
            return next();

        case RISC.WRH:
            syscall(RISC.SYS_DB_WR_REG, r[a], 0);
            return next();

        case RISC.WRL:
            syscall(RISC.SYS_IO_WR_LN, 0, 0);
            return next();

        case RISC.BREAK:
            return breakpoint();

        case RISC.SYSCALL:
            r[a] = syscall(c, r[a], r[b]);
            return next();

        default:
            return error("unknown instruction: " + toHexString(ir));
        }
    }

    //########################################################################
    // Private interface

    private String toHexString(int value) {
        return "0x" + StringHelper.toHexString(value, 8);
    }

    private IllegalArgumentException illegal(String error) {
        return new IllegalArgumentException(error);
    }

    //########################################################################
}
