package risc.emulator;

import java.io.FileReader;
import java.io.IOException;

import risc.Opcode;
import risc.RISC;
import risc.CodeParser;

import risc.syscall.Dispatcher;
import risc.syscall.Dispatchee;
import risc.syscall.InputOutput;
import risc.syscall.GarbageCollector;
import risc.syscall.Miscellaneous;
import risc.syscall.Debug;

public class Main {

    //########################################################################
    // Main interface

    public static String usage() {
        return
            "usage: java " +Emulator.class.getName()+ " [options] filename\n" +
            "\n" +
            "where options include:\n" +
            "    -mem=size  set the machine memory size to \"size\"\n" +
            "    -dump      dump memory and registers after execution\n" +
            "    -profile   show profiling statistics after execution\n" +
            "    -verbose   enable verbose output\n" +
            "    -? -help   print this help message and exit\n";
    }

    public static int run(String[] args) {
        String filename = null;
        boolean verbose = false;
        boolean dump = false;
        boolean profile = false;
        boolean help = false;
        int size = Emulator.DEFAULT_MEMORY_SIZE;

        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("-?") || args[i].equals("-help")) {
                help = true;
            } else if (args[i].equals("-dump")) {
                dump = true;
            } else if (args[i].equals("-profile")) {
                profile = true;
            } else if (args[i].equals("-verbose")) {
                verbose = true;
            } else if (args[i].startsWith("-mem=")) {
                String string = args[i].substring(5);
                try {
                    int value = Integer.parseInt(string);
                    if (value < 0) {
                        System.err.println("invalid memory size: " + value);
                        return 1;
                    } else {
                        size = value;
                    }
                } catch (NumberFormatException exception) {
                    System.err.println("invalid memory size: " + string);
                    return 1;
                }
            } else if (args[i].startsWith("-")) {
                System.err.println("unrecognized option: " + args[i]);
                System.err.print(usage());
                return 1;
            } else if (filename == null) {
                filename = args[i];
            } else {
                System.err.println("wrong argument: " + args[i]);
                System.err.print(usage());
                return 1;
            }
        }

        if (help) {
            System.out.print(usage());
            return 0;
        }

        if (filename == null) {
            System.err.print(usage());
            return 1;
        }

        Dispatchee io = new InputOutput();
        Dispatchee gc = new GarbageCollector(verbose);
        Dispatchee db = new Debug();
        Dispatchee misc = new Miscellaneous();
        Dispatcher os = new Dispatcher(new Dispatchee[] {io, gc, db, misc});
        Emulator emulator = new Emulator(size, os);

        try {
            emulator.assemble(new FileReader(filename));
        } catch (IOException exception) {
            System.err.println("IO error: " + exception);
            return -1;
        } catch (CodeParser.Exception exception) {
            System.err.println("parse error at " + exception);
            return -1;
        }

        int status = 0;
        try {
            try {
                emulator.run();
            } finally {
                emulator.syscall_nofail(RISC.SYS_IO_FLUSH, 0, 0); // !!!
            }
        } catch (Emulator.Exit exception) {
            status = exception.code;
            if (status != 0)
                System.err.println("exited with status code " + status);
            // !!! duplicated code
        } catch (Emulator.Error exception) {
            System.err.println("runtime error: " + exception.getMessage());
            status = -1;
            dump = true;
        } catch (Emulator.Exception exception) {
            System.err.println("runtime error: " + exception.toString());
            status = -1;
            dump = true;
        }

        if (dump) {
            System.out.println();
            // !!! System.out.print(emulator.dump());
            emulator.syscall_nofail(RISC.SYS_DB_DUMP, 0, 0); // !!!
        }

        if (profile) {
            System.out.println();
            System.out.println();

            System.out.println("Profiling statistics:");
            System.out.println("Opcode         Count    Cycles  Cycles %");
            System.out.println("----------------------------------------");

            long[] cntrs = emulator.getInstructionCounters();

            long totalCount = 0;
            long totalCycles = 0;

            for (int op = 0; op < RISC.opcode_count; ++op) {
                if (Opcode.codes[op] != null) {
                    totalCount += cntrs[op];
                    totalCycles += cntrs[op] * Opcode.codes[op].cost;
                }
            }

            for (int op = 0; op < RISC.opcode_count; ++op) {
                StringBuffer line = new StringBuffer();
                String opName = RISC.opcode_mnemonic[op];

                if (cntrs[op] > 0) {
                    long cycles = cntrs[op] * Opcode.codes[op].cost;
                    long percent = 100 * cycles / totalCycles;

                    line.append(opName);

                    String countS = String.valueOf(cntrs[op]);
                    padTo(line, 20 - countS.length());
                    line.append(countS);

                    String cyclesS = String.valueOf(cycles);
                    padTo(line, 30 - cyclesS.length());
                    line.append(cyclesS);

                    String percentS = String.valueOf(percent);
                    padTo(line, 40 - percentS.length());
                    line.append(percent);

                    System.out.println(line);
                }
            }

            System.out.println("----------------------------------------");
            StringBuffer line = new StringBuffer("Total:");

            String totalCountS = String.valueOf(totalCount);
            padTo(line, 20 - totalCountS.length());
            line.append(totalCountS);

            String totalCyclesS = String.valueOf(totalCycles);
            padTo(line, 30 - totalCyclesS.length());
            line.append(totalCyclesS);

            System.out.println(line);
        }

        return status;
    }

    private static void padTo(StringBuffer buf, int len) {
        int padLen = len - buf.length();
        for (int i = 0; i < padLen; ++i) buf.append(' ');
    }

    public static void main(String[] args) {
        int status = run(args);
        if (status != 0) System.exit(status);
    }

    //########################################################################
}
